"""Deserializer for an enum."""

import enum
from functools import partial
from typing import Any

import attr

from . import combinators, factories, iterator_utils
from .errors import (
    ErrorDisplayStrategy,
    InternalDeserializationError,
    InternalDeserializerFactoryError,
)
from .types import DeserializableType, Deserializer, Primitive, T


class EnumDeserializationError(InternalDeserializationError):
    """Raised when a primitive cannot be deserialized as any value of an enum."""


@attr.s(auto_attribs=True, frozen=True)
class _SingleEnumValueDeserializer(Deserializer[T]):
    enum_t: DeserializableType
    deserializer: Deserializer
    variant: Any

    def __call__(self, value: Primitive) -> T:
        v = self.deserializer(value)
        if v == self.variant.value:
            return self.variant
        raise EnumDeserializationError(
            v,
            self.variant,
            msg=f"value mismatch {repr(v)} - expected {repr(self.variant.value)}",
            display_strategy=ErrorDisplayStrategy.ALWAYS_DISPLAY,
        )


class EnumDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create an enum deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class EnumDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for an enum type."""

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create deserializer for the specified enum type."""
        if not (isinstance(type_, type) and issubclass(type_, enum.Enum)):
            raise EnumDeserializerFactoryError(type_, "not an enum type")

        deserializers, errors = iterator_utils.accumulate_errors_d(  # type: ignore[var-annotated]
            InternalDeserializerFactoryError,
            {  # type: ignore[var-annotated]
                f"variant <{variant.name}: {repr(variant.value)}>": (
                    lambda v: _SingleEnumValueDeserializer(
                        enum_t=type_,
                        deserializer=recurse_factory.create_deserializer_internal(
                            type(v.value)
                        ),
                        variant=v,
                    ),
                    variant,
                )
                for variant in list(type_)
            },
        )
        if errors:
            raise EnumDeserializerFactoryError(
                type_, EnumDeserializerFactoryError.cannot_create_msg("enum"), errors,
            )
        return combinators.OneOfDeserializer(
            deserializers, partial(EnumDeserializationError, target_t=type_)
        )



"""Deserializer for a literal."""

from functools import partial
from typing import Any

import attr
from typing_extensions import Literal

from . import combinators, factories, iterator_utils, type_utils
from .errors import (
    ErrorDisplayStrategy,
    InternalDeserializationError,
    InternalDeserializerFactoryError,
)
from .types import DeserializableType, Deserializer, Primitive, T


class LiteralDeserializationError(InternalDeserializationError):
    """Raised when a primitive cannot be deserialized as a literal."""


@attr.s(auto_attribs=True, frozen=True)
class _SingleLiteralDeserializer(Deserializer[T]):
    deserializer: Deserializer
    value: Any

    def __call__(self, value: Primitive) -> T:
        v = self.deserializer(value)
        if v == self.value:
            return v
        raise LiteralDeserializationError(
            value,
            Literal[self.value],
            msg="value mismatch - expected '{}', got '{}'".format(self.value, v),
        )


class LiteralDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a literal deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class LiteralDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for a literal type."""

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified literal type."""
        if not type_utils.get_base_of_generic_type(type_) == Literal:
            raise LiteralDeserializerFactoryError(type_, "not a literal type")

        deserializers, errors = iterator_utils.accumulate_errors_d(  # type: ignore[var-annotated]
            InternalDeserializerFactoryError,
            {
                f"variant {type_param}": (
                    lambda l: _SingleLiteralDeserializer(
                        deserializer=recurse_factory.create_deserializer_internal(
                            _get_literal_t(l)
                        ),
                        value=l,
                    ),
                    type_param,
                )
                for type_param in type_utils.get_type_parameters(type_)
            },
        )
        if errors:
            raise LiteralDeserializerFactoryError(
                type_,
                LiteralDeserializerFactoryError.cannot_create_msg("literal"),
                errors,
            )
        return combinators.OneOfDeserializer(
            deserializers,
            partial(
                LiteralDeserializationError,
                target_t=type_,
                display_strategy=ErrorDisplayStrategy.ALWAYS_DISPLAY,
            ),
        )


def _get_literal_t(type_: type) -> type:
    if type_utils.get_base_of_generic_type(type_) == Literal:
        return type_
    return type(type_)



"""Factories used to create deserializers."""

import abc
from typing import Dict, Generic, Mapping

import attr
from typing_extensions import final

from .errors import (
    DeserializationError,
    DeserializerFactoryError,
    ErrorDisplayStrategy,
    InternalDeserializationError,
    InternalDeserializerFactoryError,
)
from .types import DeserializableType, Deserializer, Primitive, T


class DeserializerFactory(abc.ABC, Generic[T]):
    """Interface for a factory class used to create Deserializers."""

    @final
    def deserialize_into(self, type_: DeserializableType, value: Primitive) -> T:
        """Deserialize a primitive as a value of the specified type."""
        try:
            return self.create_deserializer(type_)(value)
        except InternalDeserializationError as e:  # pragma: no cover
            raise DeserializationError(**vars(e))

    @final
    def create_deserializer(self, type_: DeserializableType) -> Deserializer[T]:
        """Create a deserializer for the specified type."""
        try:
            return self.create_deserializer_internal(type_)
        except InternalDeserializerFactoryError as e:  # pragma: no cover
            raise DeserializerFactoryError(**vars(e))

    @final
    def create_deserializer_internal(
        self, type_: DeserializableType
    ) -> Deserializer[T]:
        """
        Create a deserializer, recursing with this factory.

        Used internally - does not re-raise internal exceptions.
        """
        return self.create_deserializer_recursive(self, type_)

    @abc.abstractmethod
    def create_deserializer_recursive(
        self, recurse_factory: "DeserializerFactory", type_: DeserializableType
    ) -> Deserializer[T]:
        """
        Create a deserializer, recursing with the supplied factory.

        Implemented by derived classes.
        """


@attr.s(auto_attribs=True, frozen=True)
class SequenceDeserializerFactory(DeserializerFactory):
    """Deserializer factory that tries a sequence of factories in turn."""

    _deserializer_factories: Mapping[str, DeserializerFactory]

    def create_deserializer_recursive(
        self, recurse_factory: DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified type."""
        errors_: Dict[str, InternalDeserializerFactoryError] = {}
        for desc, deserializer_factory in self._deserializer_factories.items():
            try:
                return deserializer_factory.create_deserializer_recursive(
                    recurse_factory, type_
                )
            except InternalDeserializerFactoryError as e:
                errors_[desc] = e
        raise InternalDeserializerFactoryError(
            type_,
            "cannot create deserializer for type '{}'".format(get_type_name(type_)),
            errors_,
            display_strategy=ErrorDisplayStrategy.ALWAYS_DISPLAY,
        )


def get_type_name(type_: DeserializableType) -> str:
    """Return the best available user-facing name for a type."""
    return getattr(type_, "__qualname__", str(type_))


@attr.s(auto_attribs=True, frozen=True)
class CachingDeserializerFactory(DeserializerFactory):
    """Add caching function to a DeserializerFactory."""

    _deserializer_factory: DeserializerFactory
    _cache: Dict[DeserializableType, Deserializer]

    @staticmethod
    def new(deserializer_factory: DeserializerFactory) -> "CachingDeserializerFactory":
        """Create a CachingDeserializerFactory."""
        return CachingDeserializerFactory(deserializer_factory, {})

    def create_deserializer_recursive(
        self, recurse_factory: DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified type."""
        try:
            # Don't attempt to cache non-hashable types.
            hash(type_)
        except TypeError:
            return self._deserializer_factory.create_deserializer_recursive(
                recurse_factory, type_
            )
        else:
            if type_ not in self._cache:
                self._cache[
                    type_
                ] = self._deserializer_factory.create_deserializer_recursive(
                    recurse_factory, type_
                )
            return self._cache[type_]



"""Deserializer for a union of types."""

from functools import partial
from typing import Union

import attr

from . import combinators, factories, iterator_utils, pretty_printer, type_utils
from .errors import InternalDeserializationError, InternalDeserializerFactoryError
from .types import DeserializableType, Deserializer, T


class UnionDeserializationError(InternalDeserializationError):
    """Raised when a primitive cannot be deserialized as any type in a union."""


class UnionDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a union deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class UnionDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for a union of several types."""

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified union type."""
        if not type_utils.get_base_of_generic_type(type_) == Union:
            raise UnionDeserializerFactoryError(type_, "not a union type")

        deserializers, errors = iterator_utils.accumulate_errors_d(
            InternalDeserializerFactoryError,
            {
                f"variant {pretty_printer.print_type_name(type_param)}": (
                    recurse_factory.create_deserializer_internal,
                    type_param,
                )
                for type_param in type_utils.get_type_parameters(type_)
            },
        )
        if errors:
            raise UnionDeserializerFactoryError(
                type_, UnionDeserializerFactoryError.cannot_create_msg("union"), errors,
            )
        return combinators.OneOfDeserializer(
            deserializers, partial(UnionDeserializationError, target_t=type_)
        )



"""Deserializers for primitive types."""

from typing import Any, Callable, Dict, FrozenSet, List, NoReturn, Type, TypeVar, cast

import attr

from . import factories, pretty_printer, type_utils
from .errors import (
    ErrorDisplayStrategy,
    InternalDeserializationError,
    InternalDeserializerFactoryError,
)
from .types import DeserializableType, Deserializer, NotNonePrimitive, Primitive, T


class PrimitiveDeserializationError(InternalDeserializationError):
    """Raised when deserializing a primitive fails due to a type mismatch."""


class _PrimitiveMismatchError(Exception):
    pass


@attr.s(auto_attribs=True, frozen=True)
class PrimitiveDeserializer(Deserializer[T]):
    """Deserializer for a primitive type."""

    _primitive_t: DeserializableType
    _deserializer: Callable[[Primitive], object]

    def __call__(self, value: Primitive) -> T:
        """Deserialize a primitive value."""

        def get_error_msg() -> str:
            return "type mismatch - value '{}' has type '{}', expected '{}'".format(
                pretty_printer.print_primitive(value),
                pretty_printer.print_type_name(type(value)),
                pretty_printer.print_type_name(self._primitive_t),
            )

        try:
            return cast(T, self._deserializer(value))
        except _PrimitiveMismatchError:
            raise PrimitiveDeserializationError(
                value,
                self._primitive_t,
                msg=get_error_msg(),
                display_strategy=ErrorDisplayStrategy.ALWAYS_DISPLAY,
            )
        except (TypeError, ValueError) as e:
            raise PrimitiveDeserializationError(
                value, self._primitive_t, msg=get_error_msg(), cause=e
            )


class PrimitiveDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a primitive deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class PrimitiveDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for a primitive type."""

    coerce_strings: bool
    true_strings: FrozenSet[str]
    false_strings: FrozenSet[str]

    def create_deserializer_recursive(
        self, _recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified primitive type."""
        if type_ in cast(
            List[type], type_utils.get_type_parameters(Primitive)  # type: ignore[misc]
        ) + [
            Any,  # type: ignore[list-item]
            bool,
        ]:
            if self.coerce_strings:
                deserialize_int = _deserialize_int_from_string
                deserialize_float = _deserialize_float_from_string

                def inner(value: Primitive) -> bool:
                    return _deserialize_bool_from_string(
                        self.true_strings, self.false_strings, value
                    )

                deserialize_bool = inner
            else:
                deserialize_int = _deserialize_int
                deserialize_float = _deserialize_float
                deserialize_bool = _deserialize_bool

            return PrimitiveDeserializer(
                type_,
                {
                    Any: lambda x: x,
                    str: _deserialize_str,
                    int: deserialize_int,
                    float: deserialize_float,
                    dict: _deserialize_dict,
                    list: _deserialize_list,
                    bool: deserialize_bool,
                    type(None): _deserialize_none,
                }[type_],
            )
        raise PrimitiveDeserializerFactoryError(type_, "not a primitive type")


def _deserialize_str(value: Primitive) -> str:
    """Deserialise a primitive as a string."""
    return _to_primitive(value, str)


def _deserialize_int(value: Primitive) -> int:
    """
    Deserialise a primitive as an int.

    Union[bool, int] will be collapsed into int, so special-case bools.
    TODO - deserialize with best match
    """
    if isinstance(value, bool):
        return value
    return _to_primitive(value, int)


def _deserialize_int_from_string(value: Primitive) -> int:
    return int(_deserialize_str(value))


def _deserialize_float(value: Primitive) -> float:
    """
    Deserialize a primitive as a float.

    Both ints and floats are acceptable input values.
    """
    try:
        return _to_primitive(value, int)
    except _PrimitiveMismatchError:
        return _to_primitive(value, float)


def _deserialize_float_from_string(value: Primitive) -> float:
    return float(_deserialize_str(value))


def _deserialize_dict(value: Primitive) -> Dict[str, Any]:
    """Deserialize a primitive as an untyped dict."""
    return _to_primitive(value, dict)


def _deserialize_list(value: Primitive) -> List[Any]:
    """Deserialize a primitive as an untyped list."""
    return _to_primitive(value, list)


def _deserialize_bool(value: Primitive) -> bool:
    """Deserialize a primitive as a bool."""
    return _to_primitive(value, bool)


def _deserialize_bool_from_string(
    true_values: FrozenSet[str], false_values: FrozenSet[str], value: Primitive
) -> bool:
    """Deserialize a string as a bool."""
    value_str = _deserialize_str(value).lower()
    if value_str in true_values:
        return True
    elif value_str in false_values:
        return False
    else:
        raise ValueError(
            "expected one of {}, got '{}'".format(
                set(true_values | false_values), value_str
            )
        )


def _deserialize_none(value: Primitive) -> None:
    """Deserialize a primitive as None."""
    if value is None:
        return value
    _raise_type_mismatch(value, type(None))


S = TypeVar("S", bound=NotNonePrimitive)


def _to_primitive(value: Primitive, primitive_t: Type[S]) -> S:
    if isinstance(value, primitive_t):
        return cast(S, primitive_t(value))
    _raise_type_mismatch(value, primitive_t)


def _raise_type_mismatch(value: Primitive, type_: type) -> NoReturn:
    raise _PrimitiveMismatchError(
        "expected '{}', got '{}'".format(
            pretty_printer.print_type_name(type_),
            pretty_printer.print_type_name(type(value)),
        )
    )



"""Deserializer for a newtype alias."""

import attr

from . import factories
from .errors import (
    ErrorDisplayStrategy,
    InternalDeserializationError,
    InternalDeserializerFactoryError,
)
from .types import DeserializableType, Deserializer, Primitive, T


class NewTypeDeserializationError(InternalDeserializationError):
    """Raised when a primitive cannot be deserialized as a newtype."""


@attr.s(auto_attribs=True, frozen=True)
class NewTypeDeserializer(Deserializer[T]):
    """Deserializer for a newtype."""

    _newtype_t: DeserializableType
    _deserializer: Deserializer[T]

    def __call__(self, value: Primitive) -> T:
        """Deserialize a primitive as a newtype."""
        try:
            return self._deserializer(value)
        except InternalDeserializationError as e:
            raise NewTypeDeserializationError(
                value,
                self._newtype_t,
                cause=e,
                display_strategy=ErrorDisplayStrategy.ALWAYS_DISPLAY,
            )


class NewTypeDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a newtype deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class NewTypeDeserializerFactory(factories.DeserializerFactory):
    """Create a newtype deserializer."""

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified newtype."""
        if not getattr(type_, "__qualname__", None) == "NewType.<locals>.new_type":
            raise NewTypeDeserializerFactoryError(type_, "not a newtype")
        try:
            return NewTypeDeserializer(
                type_,
                recurse_factory.create_deserializer_internal(
                    type_.__supertype__  # type: ignore[union-attr]
                ),
            )
        except InternalDeserializerFactoryError as e:
            raise NewTypeDeserializerFactoryError(
                type_, NewTypeDeserializerFactoryError.cannot_create_msg("newtype"), e
            )



"""Utilities for formatting nested terramare exceptions."""

import enum
from typing import (
    Callable,
    Generic,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import attr

E = TypeVar("E", bound=Exception)
K = TypeVar("K")


Cause = Union[E, Mapping[str, E]]


@enum.unique
class ErrorDisplayStrategy(enum.Enum):
    """Strategy for displaying an exception."""

    DEFAULT = enum.auto()
    ALWAYS_DISPLAY = enum.auto()
    EAGERLY_COLLAPSE = enum.auto()


@attr.s(auto_attribs=True, frozen=True)
class ExceptionData(Generic[K, E]):
    """
    Interface for tree formatted exceptions.

    :var key: Key used for determining exception uniqueness. An exception's
        details will only be displayed once even if it occurs multiple times.
    :var cause: Cause of the exception, either another exception or a labelled
        collection of exceptions.
    :var msg: Exception message.
    :var always_display_msg: If set, the exception's message will always be
        displayed, even if it could be collapsed because either:

        - Its cause is a single exception;
        - It is part of a collection of labelled exceptions.
    """

    key: K
    cause: Cause[E]
    msg: str
    display_strategy: ErrorDisplayStrategy


def format_exception(
    exception: E, get_metadata: Callable[[E], ExceptionData[K, E]]
) -> str:
    """Recursively format a tree of exceptions."""

    def get_tree_(
        root: E, prefix: Optional[str] = None, is_root: bool = False
    ) -> _Tree[_Elt[K]]:
        """Convert an exception to a tree of messages."""

        metadata = get_metadata(root)

        def get_msg(is_leaf: bool) -> str:
            if prefix is None:
                return metadata.msg
            if (
                metadata.display_strategy == ErrorDisplayStrategy.ALWAYS_DISPLAY
                or is_leaf
                or is_root
            ):
                # Always display the message of a leaf or the root, as otherwise
                # the exception doesn't make much sense.
                return f"{prefix}: {metadata.msg}"
            return prefix

        if not isinstance(metadata.cause, Exception):
            children = cast(Mapping[str, E], metadata.cause)
            # Select "interesting" children - those that can't be eagerly
            # collapsed. If none of the children are interesting, select them
            # all.
            interesting_children = {
                key: child
                for key, child in children.items()
                if get_metadata(child).display_strategy
                != ErrorDisplayStrategy.EAGERLY_COLLAPSE
            }
            if interesting_children:
                children = interesting_children
            if len(children) != 1:
                return _Tree(
                    _Elt(metadata.key, get_msg(len(children) == 0)),
                    [get_tree_(child, f"[{key}]") for key, child in children.items()],
                )
            else:
                child = list(children.values())[0]
        else:
            child = cast(E, metadata.cause)

        if metadata.display_strategy == ErrorDisplayStrategy.ALWAYS_DISPLAY or is_root:
            return _Tree(_Elt(metadata.key, get_msg(False)), [get_tree_(child)])
        # If an exception is caused by a single other exception, and the
        # message doesn't need to be displayed, don't include that exception
        # in the tree at all. Instead, replace it by its cause.
        return get_tree_(child, prefix)

    return _format_exception_tree(get_tree_(exception, is_root=True))


@attr.s(auto_attribs=True, frozen=True)
class _Elt(Generic[K]):
    key: K
    msg: str


T = TypeVar("T")


@attr.s(auto_attribs=True, frozen=True)
class _Tree(Generic[T]):
    element: T
    children: List["_Tree[T]"]


def _format_exception_tree(root: _Tree[_Elt[K]], max_width: int = 100) -> str:
    seen: List[Tuple[str, K]] = []
    return _format_exception_tree_recursive(root, seen, max_width=max_width, depth=0)


def _format_exception_tree_recursive(
    root: _Tree[_Elt[K]], seen: List[Tuple[str, K]], max_width: int, depth: int
) -> str:
    def inner(seen: List[Tuple[str, K]]) -> str:
        if not root.children:
            return root.element.msg
        if len(root.children) == 1 and not root.children[0].children:
            # If there's only a single child, and it has no children itself,
            # display this element and its child on a single line.
            rv = root.element.msg + ": " + root.children[0].element.msg
            if 2 * depth + len(rv) <= max_width:
                # Provided the line is not too long!
                return rv
        return "\n".join(
            [root.element.msg + ":"]
            + [
                child_str
                for child_str in [
                    _indent(
                        _format_exception_tree_recursive(
                            root=child, seen=seen, max_width=max_width, depth=depth + 1
                        ),
                        "- ",
                        "  ",
                    )
                    for child in root.children
                ]
            ]
        )

    root_str = inner([])
    if (root_str, root.element.key) in seen:
        # If this exception has already been seen in the tree, don't display
        # it in full detail - just refer to the previous occurence.
        return f"{root.element.msg} as above"
    root_str = inner(seen)
    seen.append((root_str, root.element.key))
    return root_str


def _indent(s: str, head_prefix: str, tail_prefix: str) -> str:
    return "\n".join(
        [
            (head_prefix if lineno == 0 else tail_prefix) + line
            for lineno, line in enumerate(s.splitlines())
        ]
    )



"""Common types."""

import abc
from typing import Any, Callable, Generic, Type, TypeVar, Union

T = TypeVar("T")


NotNonePrimitive = Union[str, int, float, dict, list, bool]
Primitive = Union[NotNonePrimitive, None]

DeserializableType = Union[Type[Any], Callable[..., Any]]


class Deserializer(abc.ABC, Generic[T]):
    """Interface implemented by deserializers."""

    @abc.abstractmethod
    def __call__(self, value: Primitive) -> T:
        """Deserialize a value."""


class TerramareError(Exception):
    """Base class for exceptions raised by terramare."""



"""Deserializer for a (typed) tuple."""

import itertools
from functools import partial
from typing import Optional, Tuple, cast

import attr

from . import combinators, factories, iterator_utils, type_utils
from .errors import InternalDeserializationError, InternalDeserializerFactoryError
from .types import DeserializableType, Deserializer, T


class TupleDeserializationError(InternalDeserializationError):
    """Raised when a primitive cannot be deserialized as a tuple."""


class TupleDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a tuple deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class TupleDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for a tuple type."""

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified tuple type."""
        if not type_utils.get_base_of_generic_type(type_) == Tuple:
            raise TupleDeserializerFactoryError(type_, "not a tuple type")

        type_params = type_utils.get_type_parameters(type_)
        deserializers, errors = iterator_utils.accumulate_errors_l(
            InternalDeserializerFactoryError,
            zip(
                itertools.repeat(recurse_factory.create_deserializer_internal),
                (tp for tp in type_params if tp != Ellipsis),
            ),
        )
        if errors:
            raise TupleDeserializerFactoryError(
                type_,
                TupleDeserializerFactoryError.cannot_create_msg("tuple"),
                {f"at index {i}": e for i, e in errors},
            )

        var_deserializer: Optional[Deserializer] = None
        if type_params and type_params[-1] == Ellipsis:
            _, var_deserializer = deserializers[-1]
            required_deserializers = deserializers[:-1]
        else:
            required_deserializers = deserializers

        return combinators.SequenceDeserializer(
            recurse_factory.create_deserializer_internal(list),
            [ds for _, ds in required_deserializers],
            [],
            var_deserializer,
            lambda _, v: cast(T, tuple(v)),
            partial(TupleDeserializationError, target_t=type_),
        )



"""Deserializer for a generalized homogenous sequence."""

from functools import partial
from typing import FrozenSet, Iterable, Iterator, List, MutableSequence, Sequence, Set

import attr

from . import combinators, factories, pretty_printer, type_utils
from .errors import InternalDeserializationError, InternalDeserializerFactoryError
from .types import DeserializableType, Deserializer, T


class HomogenousSequenceDeserializationError(InternalDeserializationError):
    """Raised when failing to deserialize a (typed) generalized sequence."""


class HomogeneousSequenceDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a (typed) generalized sequence deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class HomogeneousSequenceDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for a generalized homogenous sequence type."""

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified homogeneous sequence type."""
        sequence_constructors = {
            FrozenSet: set,
            Iterable: list,
            Iterator: iter,
            List: list,
            MutableSequence: list,
            Sequence: list,
            Set: set,
        }
        base = type_utils.get_base_of_generic_type(type_)
        if base not in sequence_constructors:
            raise HomogeneousSequenceDeserializerFactoryError(
                type_, "not a sequence type"
            )

        [value_type] = type_utils.get_type_parameters(type_)
        try:
            deserializer = recurse_factory.create_deserializer_internal(value_type)
        except InternalDeserializerFactoryError as e:
            raise HomogeneousSequenceDeserializerFactoryError(
                type_,
                HomogeneousSequenceDeserializerFactoryError.cannot_create_msg(
                    pretty_printer.print_type_name(base)
                ),
                e,
            )

        return combinators.SequenceDeserializer(
            recurse_factory.create_deserializer_internal(list),
            [],
            [],
            deserializer,
            lambda _, v: sequence_constructors[base](v),  # type: ignore[index, operator]
            partial(HomogenousSequenceDeserializationError, target_t=type_),
        )



"""Common utility functions for dealing with iterators."""

from typing import (
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

T = TypeVar("T")


def get_single_element(l: Sequence[T]) -> Optional[T]:
    """
    Return the element contained in a single-element list.

    Return None if the list is empty, and raise an exception if more than one
    element is present.
    """
    if not l:
        return None
    if len(l) == 1:
        return l[0]
    raise ValueError("Too many elements in list: {}".format(l))  # pragma: no cover


S = TypeVar("S")


def zip_strict(
    required_ts: Sequence[T], optional_ts: Sequence[T], ss: Sequence[S]
) -> List[Tuple[T, S]]:
    """Zip two lists, raising if they are of unequal length."""
    if len(ss) < len(required_ts):
        raise ValueError(
            "too few elements ({}) - expected at least {}".format(
                len(ss), len(required_ts)
            )
        )
    ts = [*required_ts, *optional_ts]
    if len(ss) > len(ts):
        raise ValueError(
            "too many elements ({}) - expected at most {}".format(len(ss), len(ts))
        )
    return list(zip(ts, ss))


def zip_strict_extra(
    t: T, required_ts: Sequence[T], optional_ts: Sequence[T], ss: Sequence[S]
) -> List[Tuple[T, S]]:
    """
    Zip two lists, raising if the second is shorter.

    If the first is shorter, extend it with copies of t.
    """
    return zip_strict(
        required_ts,
        list(optional_ts) + [t] * (len(ss) - len(required_ts) - len(optional_ts)),
        ss,
    )


def zip_strict_dict(
    required_ts: Mapping[str, T], optional_ts: Mapping[str, T], ss: Mapping[str, S]
) -> Dict[str, Tuple[T, S]]:
    """Zip two dictionaries, raising if they do not contain the same keys."""

    def show_if(msg: str, keys: Set[str]) -> Optional[str]:
        return msg.format(", ".join(keys)) if keys else None

    missing_keys = show_if("missing keys '{}'", set(required_ts) - set(ss))
    ts = {**required_ts, **optional_ts}
    unexpected_keys = show_if("unexpected keys '{}'", set(ss) - set(ts))
    if missing_keys or unexpected_keys:
        raise ValueError(
            "key mismatch - {}".format(
                ", ".join((msg for msg in (missing_keys, unexpected_keys) if msg))
            )
        )
    return {k: (ts[k], ss[k]) for k in ss}


def zip_strict_dict_extra(
    t: T,
    required_ts: Mapping[str, T],
    optional_ts: Mapping[str, T],
    ss: Mapping[str, S],
) -> Dict[str, Tuple[T, S]]:
    """
    Zip two dictionaries, raising if the first contains keys not present in the second.

    If the second contains keys not present in the first, extend the first with copies of t.
    """
    return zip_strict_dict(
        required_ts,
        {
            **optional_ts,
            **{k: t for k in ss if k not in required_ts and k not in optional_ts},
        },
        ss,
    )


ExceptionType = Union[Type[Exception], Tuple[Type[Exception], ...]]


def accumulate_errors_l(
    exception_t: ExceptionType, fn_elts: Iterator[Tuple[Callable[[S], T], S]]
) -> Tuple[List[Tuple[int, T]], List[Tuple[int, Exception]]]:
    """Apply a function, appending the result or the raised exception to a list."""
    results, errors = accumulate_errors_d(
        exception_t, {i: fn_elt for i, fn_elt in enumerate(fn_elts)}
    )
    return list(results.items()), list(errors.items())


K = TypeVar("K")


def accumulate_errors_d(
    exception_t: ExceptionType, fn_elts: Mapping[K, Tuple[Callable[[S], T], S]]
) -> Tuple[Dict[K, T], Dict[K, Exception]]:
    """Apply a function, adding the result or the raised exception to a dict."""
    results: Dict[K, T] = {}
    errors: Dict[K, Exception] = {}
    for key, (fn, elt) in fn_elts.items():
        try:
            results[key] = fn(elt)
        except exception_t as e:
            errors[key] = e
    return results, errors


def divide_list(fn: Callable[[T], bool], ts: Sequence[T]) -> Tuple[List[T], List[T]]:
    """Divide a list into elements that meet a predicate and those that don't."""
    predicate_met: List[T] = []
    predicate_not_met: List[T] = []
    for t in ts:
        if fn(t):
            predicate_met.append(t)
        else:
            predicate_not_met.append(t)
    return (predicate_met, predicate_not_met)



"""Deserializer factory for a generic class."""

import inspect
import sys
import typing
from functools import partial
from inspect import Parameter
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    List,
    Mapping,
    Sequence,
    Type,
    TypeVar,
)

import attr

from . import combinators, factories, iterator_utils
from .errors import (
    ErrorDisplayStrategy,
    InternalDeserializationError,
    InternalDeserializerFactoryError,
)
from .types import DeserializableType, Deserializer, Primitive, T

S = TypeVar("S")


class ClassDeserializationError(InternalDeserializationError):
    """Raised when a primitive cannot be deserialized into a class."""


@attr.s(auto_attribs=True, frozen=True)
class ClassValueDeserializer(Deserializer[T]):
    """Deserialize a class from a single value."""

    _handle_exception_ts: FrozenSet[Type[Exception]]
    _fn: Callable[..., T]
    _value_deserializer: Deserializer

    def __call__(self, value: Primitive) -> T:
        """Deserialize a primitive into parameters for a function call."""
        try:
            arg = self._value_deserializer(value)
        except InternalDeserializationError as e:
            raise ClassDeserializationError(value, self._fn, cause=e)
        try:
            return self._fn(arg)
        except tuple(  # pylint: disable=catching-non-exception
            self._handle_exception_ts
        ) as e:
            raise ClassDeserializationError(value, self._fn, cause=e)


class ClassDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a class deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class ClassDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for a class."""

    _enable_if: Callable[[DeserializableType], bool] = lambda _: True
    _handle_exception_ts: FrozenSet[Type[Exception]] = attr.ib(factory=frozenset)

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified class type."""
        if not self._enable_if(type_):
            raise ClassDeserializerFactoryError(type_, "disabled for type")
        if not hasattr(type_, "__call__"):
            raise ClassDeserializerFactoryError(type_, "not a callable type")

        parameters = _get_parameters(type_)
        class_deserializers = {}
        errors = {}
        for desc, fn in {
            "from dict": _create_class_dict_deserializer,
            "from list": _create_class_list_deserializer,
            "from value": _create_class_value_deserializer,
        }.items():
            try:
                class_deserializers[desc] = fn(
                    self._handle_exception_ts, recurse_factory, type_, parameters
                )
            except InternalDeserializerFactoryError as e:
                errors[desc] = e
        if class_deserializers:
            return combinators.OneOfDeserializer(
                class_deserializers,
                partial(
                    ClassDeserializationError,
                    target_t=type_,
                    display_strategy=ErrorDisplayStrategy.ALWAYS_DISPLAY,
                ),
            )
        raise ClassDeserializerFactoryError(
            type_, ClassDeserializerFactoryError.cannot_create_msg("class"), errors
        )


def _create_class_value_deserializer(
    handle_exception_ts: FrozenSet[Type[Exception]],
    recurse_factory: factories.DeserializerFactory,
    type_: DeserializableType,
    parameters: Sequence[inspect.Parameter],
) -> Deserializer:
    required, optional = iterator_utils.divide_list(_is_required, parameters)
    if len(required) > 1 or len(required) + len(optional) < 1:
        raise ClassDeserializerFactoryError(
            type_,
            "unsupported number of parameters "
            "(required: {}, optional: {})".format(len(required), len(optional)),
        )

    value_parameter = parameters[0]
    if not _is_pos(value_parameter):
        raise ClassDeserializerFactoryError(
            type_, "non-positional first parameter {}".format(value_parameter.name),
        )
    return ClassValueDeserializer(
        handle_exception_ts,
        type_,
        _get_deserializers(recurse_factory, type_, [value_parameter])[
            value_parameter.name
        ],
    )


def _create_class_list_deserializer(
    handle_exception_ts: FrozenSet[Type[Exception]],
    recurse_factory: factories.DeserializerFactory,
    type_: DeserializableType,
    parameters: Sequence[inspect.Parameter],
) -> Deserializer:
    non_pos_required = [
        p.name for p in parameters if _is_required(p) and not _is_pos(p)
    ]
    if non_pos_required:
        raise ClassDeserializerFactoryError(
            type_,
            "required non-positional parameter(s) {}".format(
                ", ".join(non_pos_required)
            ),
        )

    pos = [p for p in parameters if _is_pos(p)]
    deserializers = _get_deserializers(recurse_factory, type_, pos)
    var_pos, non_var_pos = iterator_utils.divide_list(_is_var, pos)

    def construct_fn(value: Primitive, args: Sequence[Any]) -> T:
        try:
            return type_(*args)
        except tuple(  # pylint: disable=catching-non-exception
            handle_exception_ts
        ) as e:
            raise ClassDeserializationError(value, type_, cause=e)

    return combinators.SequenceDeserializer(
        recurse_factory.create_deserializer_internal(list),
        [deserializers[p.name] for p in non_var_pos if _is_required(p)],
        [deserializers[p.name] for p in non_var_pos if not _is_required(p)],
        iterator_utils.get_single_element([deserializers[p.name] for p in var_pos]),
        construct_fn,
        partial(ClassDeserializationError, target_t=type_),
    )


def _create_class_dict_deserializer(
    handle_exception_ts: FrozenSet[Type[Exception]],
    recurse_factory: factories.DeserializerFactory,
    type_: DeserializableType,
    parameters: Sequence[inspect.Parameter],
) -> Deserializer:
    non_kw_required = [p.name for p in parameters if _is_required(p) and not _is_kw(p)]
    if non_kw_required:  # pragma: no cover TODO
        raise ClassDeserializerFactoryError(
            type_,
            "required non-keyword parameter(s) {}".format(", ".join(non_kw_required)),
        )

    kw = [p for p in parameters if _is_kw(p)]
    deserializers = _get_deserializers(recurse_factory, type_, kw)
    var_kw, non_var_kw = iterator_utils.divide_list(_is_var, kw)

    def construct_fn(value: Primitive, args: Mapping[str, Any]) -> T:
        try:
            return type_(**args)
        except tuple(  # pylint: disable=catching-non-exception
            handle_exception_ts
        ) as e:
            raise ClassDeserializationError(value, type_, cause=e)

    return combinators.MappingDeserializer(
        recurse_factory.create_deserializer_internal(dict),
        recurse_factory.create_deserializer_internal(str),
        {p.name: deserializers[p.name] for p in non_var_kw if _is_required(p)},
        {p.name: deserializers[p.name] for p in non_var_kw if not _is_required(p)},
        iterator_utils.get_single_element([deserializers[p.name] for p in var_kw]),
        construct_fn,
        partial(ClassDeserializationError, target_t=type_),
    )


def _get_parameters(type_: DeserializableType) -> List[inspect.Parameter]:
    try:
        return list(inspect.signature(type_).parameters.values())
    except ValueError as e:
        raise ClassDeserializerFactoryError(
            type_,
            ClassDeserializerFactoryError.cannot_create_msg("class"),
            e,
            display_strategy=ErrorDisplayStrategy.ALWAYS_DISPLAY,
        )


def _get_deserializers(
    recurse_factory: factories.DeserializerFactory,
    type_: DeserializableType,
    parameters: List[inspect.Parameter],
) -> Dict[str, Deserializer]:
    type_hints = _get_type_hints(type_, parameters)
    deserializers, errors = iterator_utils.accumulate_errors_d(
        InternalDeserializerFactoryError,
        {
            parameter.name: (
                lambda p: recurse_factory.create_deserializer_internal(
                    type_hints[p.name]
                ),
                parameter,
            )
            for parameter in parameters
        },
    )

    if errors:
        raise ClassDeserializerFactoryError(
            type_,
            ClassDeserializerFactoryError.cannot_create_msg("class"),
            {f'parameter "{k}"': e for k, e in errors.items()},
        )
    return deserializers


def _get_type_hints(
    type_: DeserializableType, params: Sequence[Parameter]
) -> Dict[str, Any]:
    if hasattr(type_, "__globals__"):
        globals_ = getattr(type_, "__globals__")
    elif hasattr(type_, "__module__"):
        globals_ = vars(sys.modules[getattr(type_, "__module__")])
    else:  # pragma: no cover TODO
        globals_ = {}

    # For some reason `get_type_hints` can't handle the type returned by
    # `functools.partial` - but it _can_ handle its `__call__` method.
    if isinstance(type_, partial):
        type_ = type_.__call__

    try:
        type_hints = typing.get_type_hints(type_, globals_)
    except TypeError as e:  # pragma: no cover
        raise ClassDeserializerFactoryError(
            type_, ClassDeserializerFactoryError.cannot_create_msg("class"), e
        )

    def get_type_hint(p: inspect.Parameter) -> Any:
        if p.name in type_hints:
            return type_hints[p.name]
        if p.annotation == inspect.Parameter.empty:
            return Any
        if not isinstance(p.annotation, str):
            return p.annotation
        # TODO __closure__?
        return eval(  # pragma: no cover, pylint: disable=eval-used
            p.annotation, globals_
        )

    return {p.name: get_type_hint(p) for p in params}


def _is_var(parameter: inspect.Parameter) -> bool:
    return parameter.kind in {Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD}


def _is_pos(parameter: inspect.Parameter) -> bool:
    return parameter.kind in {
        Parameter.POSITIONAL_ONLY,
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.VAR_POSITIONAL,
    }


def _is_kw(parameter: inspect.Parameter) -> bool:
    return parameter.kind in {
        Parameter.KEYWORD_ONLY,
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.VAR_KEYWORD,
    }


def _is_required(parameter: inspect.Parameter) -> bool:
    return parameter.default == inspect.Parameter.empty and not _is_var(parameter)



"""Deserializer combinators."""

import itertools
from functools import partial
from typing import (
    Any,
    Callable,
    Generic,
    Hashable,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
)

import attr
from typing_extensions import Protocol

from . import exception_formatter, iterator_utils
from .errors import ErrorDisplayStrategy, InternalDeserializationError
from .types import Deserializer, Primitive, T


class _ErrorT(Protocol):
    def __call__(
        self,
        value: Primitive,
        cause: exception_formatter.Cause[Exception],
        display_strategy: ErrorDisplayStrategy = ErrorDisplayStrategy.DEFAULT,
    ) -> InternalDeserializationError:  # pragma: no cover
        ...


@attr.s(auto_attribs=True, frozen=True)
class OneOfDeserializer(Deserializer[T]):
    """Deserialize with one of the supplied set of deserializers."""

    _deserializers: Mapping[str, Deserializer]
    _error_t: _ErrorT

    def __call__(self, value: Primitive) -> T:
        """Deserialize a primitive by trying several deserializers in turn."""
        errors_ = {}
        for desc, ds in self._deserializers.items():
            try:
                return ds(value)
            except InternalDeserializationError as e:
                errors_[desc] = e
        raise self._error_t(
            value, cause=errors_ if len(errors_) > 1 else list(errors_.values())[0],
        )


@attr.s(auto_attribs=True, frozen=True)
class SequenceDeserializer(Deserializer[T]):
    """Deserialize a list using a sequence of deserializers."""

    _untyped_list_deserializer: Deserializer[list]

    _required_elt_deserializers: Sequence[Deserializer]
    _optional_elt_deserializers: Sequence[Deserializer]
    _var_elt_deserializer: Optional[Deserializer]

    _construct_fn: Callable[[Primitive, Sequence[Any]], T]
    _error_t: _ErrorT

    def __call__(self, value: Primitive) -> T:
        """
        Deserialize a primitive using a sequence of deserializers.

        Each list element is matched with a deserializer based on its index.
        From the start of the list:
        - Elements are first matched with the required element deserializers,
          raising an error if there are too few elements;
        - Additional elements are then matched with the optional element
          deserializers;
        - Any remaining elements are matched with the var element deserializer,
          if supplied. Otherwise, an error is raised.
        """
        try:
            elts = self._untyped_list_deserializer(value)
        except InternalDeserializationError as e:
            raise self._error_t(
                value, cause=e, display_strategy=ErrorDisplayStrategy.EAGERLY_COLLAPSE
            )

        try:
            zipped = (
                partial(  # type: ignore[operator]
                    iterator_utils.zip_strict_extra, self._var_elt_deserializer
                )
                if self._var_elt_deserializer
                else iterator_utils.zip_strict
            )(self._required_elt_deserializers, self._optional_elt_deserializers, elts,)
        except ValueError as e:
            raise self._error_t(value, cause=e)

        results, errors = iterator_utils.accumulate_errors_l(
            InternalDeserializationError, iter(zipped),
        )

        if errors:
            raise self._error_t(
                value, cause={f"at index {i}": e for i, e in errors},
            )
        return self._construct_fn(value, [v for _, v in results])


K = TypeVar("K", bound=Hashable)


@attr.s(auto_attribs=True, frozen=True)
class MappingDeserializer(Generic[K, T], Deserializer[T]):
    """Deserialize a dictionary using a mapping of keys to deserializers."""

    _untyped_dict_deserializer: Deserializer[dict]

    _key_deserializer: Deserializer[K]
    _required_elt_deserializers: Mapping[K, Deserializer]
    _optional_elt_deserializers: Mapping[K, Deserializer]
    _var_elt_deserializer: Optional[Deserializer]

    _construct_fn: Callable[[Primitive, Mapping[K, Any]], T]
    _error_t: _ErrorT

    def __call__(self, value: Primitive) -> T:
        """
        Deserialize a primitive using a mapping of keys to deserializers.

        Each dictionary element is matched with a deserializer based on its key.
        An error will be raised if the dictionary is missing keys corresponding
        to required element deserializers.
        An error will also be raised if the dictionary contains keys that do not
        correspond to either required or optional element deserializers, and
        no var element deserializer is supplied.
        """
        try:
            elts = self._untyped_dict_deserializer(value)
        except InternalDeserializationError as e:
            raise self._error_t(
                value, cause=e, display_strategy=ErrorDisplayStrategy.EAGERLY_COLLAPSE
            )

        keys, key_errors = iterator_utils.accumulate_errors_l(
            InternalDeserializationError,
            zip(itertools.repeat(self._key_deserializer), elts.keys()),
        )

        if key_errors:
            raise self._error_t(
                value, cause={f"at index {i}": e for i, e in key_errors}
            )

        elts = dict(zip((k for _, k in keys), elts.values()))

        try:
            zipped = (
                partial(  # type: ignore[operator]
                    iterator_utils.zip_strict_dict_extra, self._var_elt_deserializer
                )
                if self._var_elt_deserializer
                else iterator_utils.zip_strict_dict
            )(self._required_elt_deserializers, self._optional_elt_deserializers, elts,)
        except ValueError as e:
            raise self._error_t(value, cause=e)

        results, errors = iterator_utils.accumulate_errors_d(
            InternalDeserializationError, zipped
        )

        if errors:
            raise self._error_t(
                value, cause={f'at key "{k}"': e for k, e in errors.items()},
            )
        return self._construct_fn(value, results)



"""Automatically deserialize complex objects from simple Python types."""

from typing import Callable, FrozenSet, Type

from . import (
    classes,
    dicts,
    enums,
    errors,
    factories,
    literals,
    newtypes,
    primitives,
    sequences,
    tuples,
    unions,
)
from .types import DeserializableType, Primitive, T

DEFAULT_TRUE_STRINGS = frozenset({"yes", "on", "true", "1"})
DEFAULT_FALSE_STRINGS = frozenset({"no", "off", "false", "0"})
DEFAULT_EXCEPTION_TYPES_TO_HANDLE: FrozenSet[Type[Exception]] = frozenset()


def deserialize_into(
    type_: DeserializableType,
    value: Primitive,
    coerce_strings: bool = False,
    true_strings: FrozenSet[str] = DEFAULT_TRUE_STRINGS,
    false_strings: FrozenSet[str] = DEFAULT_FALSE_STRINGS,
    handle_exception_types: FrozenSet[
        Type[Exception]
    ] = DEFAULT_EXCEPTION_TYPES_TO_HANDLE,
    handle_reentrancy: bool = False,
) -> T:
    """
    Deserialize a primitive as a value of the specified type.

    :param `type_`: Deserialize into this type.
    :param `value`: Primitive value to attempt to deserialize.
    :param `coerce_strings`: If set, attempt to convert :python:`str` values to
        :python:`bool`, :python:`int`, or :python:`float` where the latter are required.
        For example:

        >>> deserialize_into(int, "1")
        Traceback (most recent call last):
            ...
        terramare.errors.DeserializationError: ...
        >>> deserialize_into(int, "1", coerce_strings=True)
        1

        This option defaults to :python:`False`.

    :param `true_strings`: Set of strings to convert to :python:`True` when convering a
        :python:`str` value to a :python:`bool`. Case is ignored.

        This value defaults to :python:`{"yes", "on", "true", "1"}`.

    :param `false_strings`: Set of strings to convert to :python:`False` when convering
        a :python:`str` value to a :python:`bool`. Case is ignored.

        This value defaults to :python:`{"no", "off", "false", "0"}`.

    :param `handle_exception_types`: Set of additional exception types that
        :python:`terramare` should catch and handle rather than propogating.

        Generally this will still result in an exception being raised. However, it will
        be a :python:`terramare` exception containing additional context.

        This option is useful when the deserialization target provides some form of
        additional validation. For example:

        >>> from typing import Union
        >>> import attr
        >>>
        >>> @attr.s
        ... class Paint:
        ...    color: str = attr.ib(validator=attr.validators.in_(["red", "blue"]))

        When no exception types are provided, context is lost when deserializing into a
        single type:

        >>> deserialize_into(Paint, "green")
        Traceback (most recent call last):
        ...
        ValueError: 'color' must be in ['red', 'blue'] (got 'green')

        When the exception type is provided, that context is preserved:

        >>> deserialize_into(Paint, "green", handle_exception_types={ValueError})
        Traceback (most recent call last):
        ...
        terramare.errors.DeserializationError: cannot read value '"green"' into 'Paint': 'color' must be in ['red', 'blue'] (got 'green')

        The exception is when the deserialization target is a union type, where
        :python:`terramare` will continue to try further union variants - as it would
        when encountering a deserialization failure.

        When no exception types are provided, :python:`terramare` cannot try further
        union variants if one fails due to a validation error:

        >>> deserialize_into(Union[Paint, str], "green")
        Traceback (most recent call last):
        ...
        ValueError: 'color' must be in ['red', 'blue'] (got 'green')

        When the exception type is provided, :python:`terramare` continues and
        successfully deserializes into a later variant.

        >>> deserialize_into(Union[Paint, str], "green", handle_exception_types={ValueError})
        'green'

    :param `handle_reentrancy`: This option determines :python:`terramare`'s behaviour
        when deserializing into a class or function that itself invokes
        :python:`terramare`:

        - If set to :python:`True`, :python:`terramare` will catch and handle exceptions
          raised by calls back into :python:`terramare`. This setting is useful when
          deserializing in several stages, common when the type of some later member
          depends on the value of an earlier one.

        - If set to :python:`False`, :python:`terramare` will not catch these exceptions,
          propogating them as normal. This is a safer setting when the expectation is
          that deserializion will not have multiple stages, that is, when
          :python:`terramare` exceptions raised by re-entrancy are unexpected and should
          be propogated.

        For example:

        >>> from typing import Any, List, Union
        >>> import attr
        >>>
        >>> @attr.s(auto_attribs=True)
        ... class PythonConfig:
        ...    pip_path: str
        >>>
        >>> @attr.s(auto_attribs=True)
        ... class BashConfig:
        ...     set_options: List[str]
        >>>
        >>> def load_language_config(
        ...     language: str,
        ...     **remainder: Any
        ... ) -> "Union[PythonConfig, BashConfig]":
        ...     if language == "python":
        ...         return deserialize_into(PythonConfig, remainder)
        ...     elif language == "bash":
        ...         return deserialize_into(PythonConfig, remainder)
        ...     else:
        ...         raise NotImplementedError
        >>>
        >>> data = {"language": "python", "set_options": "xe"}

        When handling re-entrancy is disabled, the :python:`load_language_context` is
        lost:

        >>> deserialize_into(load_language_config, data)
        Traceback (most recent call last):
        ...
        terramare.errors.DeserializationError: cannot read value ... into 'PythonConfig':
        ...

        When handling re-entrancy is enabled, that context is preserved:

        >>> deserialize_into(load_language_config, data, handle_reentrancy=True)
        Traceback (most recent call last):
        ...
        terramare.errors.DeserializationError: cannot read value ... into 'load_language_config':
        ...

    :raises terramare.DeserializerFactoryError: if a deserializer for :python:`type_`
        cannot be created.
    :raises terramare.DeserializationError: if the deserializer fails to deserialize a
        value of :python:`type_` from :python:`value`.
    """  # noqa: F401, E501
    return create_deserializer_factory(
        coerce_strings=coerce_strings,
        true_strings=true_strings,
        false_strings=false_strings,
        handle_exception_types=handle_exception_types,
        handle_reentrancy=handle_reentrancy,
    ).deserialize_into(type_, value)


def create_deserializer_factory(
    coerce_strings: bool = False,
    true_strings: FrozenSet[str] = DEFAULT_TRUE_STRINGS,
    false_strings: FrozenSet[str] = DEFAULT_FALSE_STRINGS,
    handle_exception_types: FrozenSet[
        Type[Exception]
    ] = DEFAULT_EXCEPTION_TYPES_TO_HANDLE,
    handle_reentrancy: bool = False,
) -> factories.DeserializerFactory:
    """Create a DeserializerFactory using sensible defaults which may be overridden."""

    def class_deserializer_enable_if(t: DeserializableType) -> bool:
        return getattr(t, "__origin__", None) is not Callable

    return factories.CachingDeserializerFactory.new(
        factories.SequenceDeserializerFactory(
            {
                "newtype": newtypes.NewTypeDeserializerFactory(),
                "primitive": primitives.PrimitiveDeserializerFactory(
                    coerce_strings=coerce_strings,
                    true_strings=frozenset(s.lower() for s in true_strings),
                    false_strings=frozenset(s.lower() for s in false_strings),
                ),
                "literal": literals.LiteralDeserializerFactory(),
                "tuple": tuples.TupleDeserializerFactory(),
                "union": unions.UnionDeserializerFactory(),
                "sequences": sequences.HomogeneousSequenceDeserializerFactory(),
                "dict": dicts.DictDeserializerFactory(),
                "enum": enums.EnumDeserializerFactory(),
                "class": classes.ClassDeserializerFactory(
                    enable_if=class_deserializer_enable_if,
                    handle_exception_ts=handle_exception_types
                    | frozenset(
                        {errors.DeserializationError} if handle_reentrancy else {}
                    ),
                ),
            }
        )
    )



"""Common error types."""

from typing import Optional, TypeVar, cast

import attr
from typing_extensions import final

from . import exception_formatter, pretty_printer
from .exception_formatter import ErrorDisplayStrategy
from .types import DeserializableType, Primitive, TerramareError

S = TypeVar("S", bound="DeserializationError")


@attr.s(auto_attribs=True, frozen=True)
class DeserializationError(TerramareError):
    """Raised when a deserializer fails."""

    value: Primitive
    target_t: DeserializableType
    cause: Optional[exception_formatter.Cause[Exception]] = None
    msg: Optional[str] = None
    display_strategy: ErrorDisplayStrategy = ErrorDisplayStrategy.DEFAULT

    @final
    def _get_msg(self) -> str:
        """Return a string description of the error."""
        if self.msg is not None:
            return self.msg
        return "cannot read value '{}' into '{}'".format(
            pretty_printer.print_primitive(self.value),
            pretty_printer.print_type_name(self.target_t),
        )

    @final
    def __str__(self) -> str:
        """Format the exception message."""

        def get_metadata(
            e: Exception,
        ) -> exception_formatter.ExceptionData[Exception, Exception]:
            if isinstance(e, DeserializationError):
                return exception_formatter.ExceptionData(
                    e,
                    e.cause if e.cause is not None else {},
                    e._get_msg(),  # pylint: disable=protected-access
                    e.display_strategy,
                )
            return exception_formatter.ExceptionData(
                e, {}, str(e), ErrorDisplayStrategy.DEFAULT
            )

        return exception_formatter.format_exception(cast(Exception, self), get_metadata)


class InternalDeserializationError(DeserializationError):
    """Raised when a deserializer fails. Caught internally."""


T = TypeVar("T", bound="DeserializerFactoryError")


@attr.s(auto_attribs=True, frozen=True)
class DeserializerFactoryError(TerramareError):
    """Raised when deserializer creation fails."""

    target_t: DeserializableType
    msg: str
    cause: Optional[exception_formatter.Cause[Exception]] = None
    display_strategy: ErrorDisplayStrategy = ErrorDisplayStrategy.DEFAULT

    @final
    @staticmethod
    def cannot_create_msg(deserializer_desc: str) -> str:
        """Format the message for standard deserializer creation failure."""
        return "cannot create {} deserializer".format(deserializer_desc)

    @final
    def __str__(self) -> str:
        """Format the exception message."""

        def get_metadata(
            e: Exception,
        ) -> exception_formatter.ExceptionData[Exception, Exception]:
            if isinstance(e, DeserializerFactoryError):
                return exception_formatter.ExceptionData(
                    e,
                    e.cause if e.cause is not None else {},
                    e.msg,
                    e.display_strategy,
                )
            return exception_formatter.ExceptionData(
                e, {}, str(e), ErrorDisplayStrategy.DEFAULT
            )

        return exception_formatter.format_exception(cast(Exception, self), get_metadata)


class InternalDeserializerFactoryError(DeserializerFactoryError):
    """Raised when deserializer creation fails. Caught internally."""



"""Deserializer for a dictionary."""

from functools import partial
from typing import Dict, List, Mapping, MutableMapping, Set, cast

import attr

from . import combinators, factories, iterator_utils, pretty_printer, type_utils
from .errors import InternalDeserializationError, InternalDeserializerFactoryError
from .types import DeserializableType, Deserializer, T


class DictDeserializationError(InternalDeserializationError):
    """Raised when a primitive cannot be deserialized as a dict."""


class DictDeserializerFactoryError(InternalDeserializerFactoryError):
    """Raised when failing to create a dict deserializer."""


@attr.s(auto_attribs=True, frozen=True)
class DictDeserializerFactory(factories.DeserializerFactory):
    """Create deserializer for a dictionary of values of the same type."""

    def create_deserializer_recursive(
        self, recurse_factory: factories.DeserializerFactory, type_: DeserializableType
    ) -> Deserializer[T]:
        """Create a deserializer for the specified dictionary type."""
        if type_utils.get_base_of_generic_type(type_) not in {
            Dict,
            Mapping,
            MutableMapping,
        }:
            raise DictDeserializerFactoryError(type_, "not a dict type")

        key_type, value_type = type_utils.get_type_parameters(type_)
        if not _is_hashable(key_type):
            raise DictDeserializerFactoryError(
                type_,
                "unhashable key type: '{}'".format(
                    pretty_printer.print_type_name(key_type)
                ),
            )

        deserializers, errors = iterator_utils.accumulate_errors_d(
            InternalDeserializerFactoryError,
            {
                "key": (recurse_factory.create_deserializer_internal, key_type),
                "value": (recurse_factory.create_deserializer_internal, value_type),
            },
        )
        if errors:
            raise DictDeserializerFactoryError(
                type_,
                DictDeserializerFactoryError.cannot_create_msg("dict"),
                {f"{k} type": e for k, e in errors.items()},
            )
        return combinators.MappingDeserializer(
            recurse_factory.create_deserializer_internal(dict),
            deserializers["key"],
            {},
            {},
            deserializers["value"],
            lambda _, v: cast(T, dict(v)),
            partial(DictDeserializationError, target_t=type_),
        )


def _is_hashable(type_: DeserializableType) -> bool:
    return (
        "__hash__" in dir(type_)
        and "__eq__" in dir(type_)
        and type_utils.get_base_of_generic_type(type_) not in {Dict, List, Set}
    )
